import functools
import types

import more_itertools
import torch
import torch.nn as nn
import torch.optim

# This kernel combines a list of modules into a sequence.
# Please ensure elements of list have compatible in/out sizes.
# This is necessary over raw nn.Sequential(...) since I 
# need to save/restore hidden states. 
class SequentialKernel(nn.Module):
    def __init__(self, neural_modules):
        super(SequentialKernel, self).__init__()
        self.__neural_modules_list = list(more_itertools.always_iterable(neural_modules))
        self.input_dimensions = self.__neural_modules_list[0].input_dimensions
        self.output_dimension = self.__neural_modules_list[-1].output_dimension

        self.neural_modules = nn.Sequential(*self.__neural_modules_list)

                
    def recurrent(self):
        for mod in self.__neural_modules_list:
            if mod.recurrent(): return True
        else: return False

    def save_hidden(self):
        assert self.recurrent()
        return {id(self.mod): mod.save_hidden() for mod in self.__neural_modules_list if mod.recurrent()}
            

    def restore_hidden(self, state_dict=None):
        for mod in self.__neural_modules_list:
            if mod.recurrent():
                if state_dict != None and id(mod) in state_dict: mod.restore_hidden(state_dict[id(mod)])

    def forward(self, input):
        return self.neural_modules(input)



import functools

import more_itertools
import torch
import torch.nn as nn

import librl.agent
# The random agent random selects one edge pair to toggle per timestep.
@librl.agent.add_agent_attr()
class RandomAgent(nn.Module):
    def __init__(self, observation_space, action_space):
        # Must initialize torch.nn.Module
        super(RandomAgent, self).__init__()
        self.action_space = action_space
        self.input_dimension = list(more_itertools.always_iterable(observation_space.shape))
        self.__input_size = functools.reduce(lambda x,y: x*y, self.input_dimension, 1)

    # Our action is just asking the pytorch implementation for a random set of nodes.
    def act(self, inputs):
        return self.forward(inputs)

    # Implement required pytorch interface
    def forward(self, adj):
        count = adj.view(-1, self.__input_size).shape[0]
        actions = [self.action_space.sample() for _ in range(count)]
        randoms = torch.tensor(actions, device=adj.device) # type: ignore
        # Currently makes little sense to ask "what was the probability of drawing this random action"
        # with gym spaces. TODO: Make log prob a real, useful number.
        return randoms, torch.full((count,), float('nan')) # type: ignore

    # Random agent has no parameters.
    def steal(self):
        return {}
    def stuff(self, vals):
        pass


import collections
import functools
import types

import more_itertools
import torch
import torch.nn as nn
import torch.optim

# Determine the size of a dimension after applying a pool / convolutional layer.
def resize_convolution(x, kernel_size, dilation, stride, padding):
    x = int(1 + (x + 2*padding - dilation * (kernel_size - 1) - 1)/stride)
    return x

# Determine the size of a dimension after applying a transposed convolution layer.
def resize_transpose_convolution(x, kernel_size, dilation, stride, padding, output_padding):
    t1 = (x-1)*stride
    t2 = 2*padding
    t3 = dilation*(kernel_size-1)
    t4 = output_padding
    return t1 - t2 + t3 + t4 + 1
    
# Class containing overlapping parameters been convolutional, pooling layers
class conpool_core:
    def __init__(self, kernel, stride, padding, dilation, non_linear_after):
        self.kernel = kernel
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.non_linear_after = non_linear_after

# Class describing a 2D convolutional layer.
class conv_def(conpool_core):
    def __init__(self, kernel, out_channels, stride=1, padding=0, dilation=1, non_linear_after=True, padding_type='zeros'):
        super(conv_def, self).__init__(kernel, stride, padding, dilation, non_linear_after)
        self.out_channels = out_channels
        self.padding_type = padding_type

# Class describing a transposed 2d convolutional layer.
class conv_transpose_def(conpool_core):
    def __init__(self, kernel, out_channels, stride=1, padding=0, dilation=1, output_padding=0, non_linear_after=True):
        super(conv_transpose_def, self).__init__(kernel, stride, padding, dilation, non_linear_after)
        self.out_channels = out_channels
        self.output_padding = output_padding
        self.padding_type = 'zeros'

# Class describing a 2D pooling layer.
class pool_def(conpool_core):
    def __init__(self, kernel, stride=None, padding=0, dilation=1, non_linear_after=False, pool_type='avg'):
        super(pool_def, self).__init__(kernel, stride, padding, dilation, non_linear_after)
        if stride is None:
            self.stride = self.kernel
        self.pool_type = pool_type
def construct_layers(conv_layers, input_dimensions, input_channels, dropout, dims=2):
        assert 1 <= dims <= 2
        conv_ctor, convt_ctor, pool_ctor = None, None, {}

        if dims == 1:
            conv_ctor = nn.Conv1d
            convt_ctor = nn.ConvTranspose1d
            pool_ctor['max'] = nn.MaxPool1d
            pool_ctor['avg'] = nn.AvgPool1d
        if dims == 2:
            conv_ctor = nn.Conv2d
            convt_ctor = nn.ConvTranspose2d
            pool_ctor['max'] = nn.MaxPool2d
            pool_ctor['avg'] = nn.AvgPool2d

        # Construct convolutional layers.
        curret_dims = list(input_dimensions)
        conv_list = []
        non_linear = torch.nn.LeakyReLU()
        nlo_name = "ReLU"
        output_channels = input_channels
        output_dimension = (*input_dimensions, input_channels)

        # Iterate over all pooling/convolutional layer configurations.
        # Construct all items as a (name, layer) tuple so that the layers may be loaded into
        # an ordered dictionary. Ordered dictionaries respect the order in which items were inserted,
        # and are the least painful way to construct a nn.Sequential object.
        for index, item in enumerate(conv_layers):
            # Next item is a convolutional layer, so construct one and re-compute H,W, channels.
            if isinstance(item, conv_def):
                conv_list.append((f'conv{index}', conv_ctor(output_channels, item.out_channels, item.kernel,
                    stride=item.stride, padding=item.padding, dilation=item.dilation, padding_mode=item.padding_type)))
                for dim, val in enumerate(curret_dims):
                    curret_dims[dim] = resize_convolution(val, item.kernel, item.dilation, item.stride, item.padding)
                output_channels = item.out_channels
            # Next item is a transposed convolutional layer.
            if isinstance(item, conv_transpose_def):
                conv_list.append((f'conv{index} transpose', convt_ctor(input_channels, item.out_channels, item.kernel,
                    stride=item.stride, padding=item.padding, dilation=item.dilation, padding_mode=item.padding_type, output_padding=item.output_padding)))
                for dim, val in enumerate(curret_dims):
                    curret_dims[dim] = resize_convolution(val, item.kernel, item.dilation, item.stride, item.padding)
                output_channels = item.out_channels
            # Next item is a pooling layer, so construct one and re-compute H,W.
            elif isinstance(item, pool_def):
                if item.pool_type.lower() == 'avg':
                    conv_list.append((f'avgpool{index}', pool_ctor['avg'](item.kernel, stride=item.stride, padding=item.padding)))
                    assert item.dilation == 1
                elif item.pool_type.lower() == 'max':
                    conv_list.append((f'maxpool{index}', pool_ctor['max'](item.kernel, stride=item.stride, padding=item.padding, dilation=item.dilation)))
                else:
                    raise NotImplementedError(f"{item.pool_type.lower()} is not an implemented form of pooling.")
                for dim, val in enumerate(curret_dims):
                    curret_dims[dim] = resize_convolution(val, item.kernel, item.dilation, item.stride, item.padding)

            output_dimension = (output_channels, *curret_dims)
            # Add a non-linear operator if specified by item. Non linear operators also pair with dropout
            # in all the examples I've seen
            if item.non_linear_after:
                conv_list.append((f"{nlo_name}{index}", non_linear))
                conv_list.append((f"dropout{index}", nn.Dropout(dropout)))
            #if print_sizes: print(f"Layer {index} is ({H} x {W})")
        return conv_list, output_dimension

class ConvolutionalKernel(nn.Module):
    def __init__(self, conv_layers, input_dimensions, input_channels, dropout=0, dims=2):
        super(ConvolutionalKernel, self).__init__()
        assert len(input_dimensions) == dims
        
        conv_list, out_dims = construct_layers(conv_layers, input_dimensions, input_channels, dropout, dims=dims)
        self.conv_layers = nn.Sequential(collections.OrderedDict(conv_list))
        self.input_dimensions = (input_channels, *input_dimensions)
        self.__input__size = functools.reduce(lambda x, y: x*y, self.input_dimensions, 1)
        self.output_dimension = out_dims

        # Randomize initial parameters
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.kaiming_normal_(p)
                
    def recurrent(self):
        return False

    def forward(self, input):
        # Magic line of code needed to cast image vector into correct dimensions?
        input = input.view(-1, *self.input_dimensions)
        output = self.conv_layers(input)
        
        assert not torch.isnan(output).any() # type: ignore
        return output


import functools

import librl.agent.pg, librl.agent.mdp
import librl.task
import librl.train.cc

@functools.singledispatch
def policy_gradient_update(agent, episode_iterable):
    # If you introduce a new agent type, and you wish to allow this agent to be updated via
    # policy gradient, you must register an override for this function.
    raise NotImplemented(f"Policy gradient update not implemented for this agent type.")

@policy_gradient_update.register(librl.agent.mdp.RandomAgent)
def _(agent, tasks_iterable):
    pass

@policy_gradient_update.register(librl.agent.pg.REINFORCEAgent)
def _(agent, tasks_iterable):
    librl.train.cc.update_unshared_cc_agent(agent, tasks_iterable, 1, 'actor_loss', -1, 'actor_optimizer')

@policy_gradient_update.register(librl.agent.pg.ActorCriticAgent)
def _(agent, tasks_iterable):
    librl.train.cc.update_unshared_cc_agent(agent, tasks_iterable, agent.critic_steps, 'critic_loss', 1, 'critic_optimizer')
    librl.train.cc.update_unshared_cc_agent(agent, tasks_iterable, 1, 'actor_loss', -1, 'actor_optimizer')

# Assumes there is no interaction between different agents.
# If components of an agent are shared between tasks, gradient updates are no longer independent.
def policy_gradient_step(task_samples):
    # Determine which agents are present in this training run.
    agents = { (id(task.agent),task.agent) for task in task_samples}
    #for _, agent in agents: 
        #if agent.recurrent(): agent.restore_hidden()
    for task in task_samples:
        assert task.problem_type == librl.task.ProblemTypes.ContinuousControl
        task.sample(task)

    # Collate tasks by the agent running them.
    collated_tasks = {id_agent:[] for id_agent, _ in agents}
    for task in task_samples: collated_tasks[id(task.agent)].append(task)
    for id_agent, agent in agents: policy_gradient_update(agent, collated_tasks[id_agent])


import torch

import librl.task



# Given a set of classification task samples, train the task's classifer on the task data.
# This helper method assumes classifiers do not share components, and that we are
# only dealing in assigning a single label to each data element.
def train_single_label_classifier(task_samples):
    classifiers = { (id(task.classifier),task.classifier) for task in task_samples}
    correct, total = [], []
    for task in task_samples:
        assert task.problem_type == librl.task.ProblemTypes.Classification
        for dataloader in task.train_data_iter:
            task.classifier.train()
            for batch_idx, (data, target) in enumerate(dataloader):
                loss, selected = task.train_batch(task.classifier, task.criterion, data, target)
        local_correct, local_total = 0,0
        for dataloader in task.validation_data_iter:
            task.classifier.eval()
            for batch_idx, (data, target) in enumerate(dataloader):
                loss, selected = task.validate_batch(task.classifier, task.criterion, data, target)
                local_correct += torch.eq(selected, target).sum() 
                local_total += float(target.shape[0])
        correct.append(local_correct), total.append(local_total)
    return correct, total
       


import functools

import more_itertools
import torch
import torch.nn as nn
import torch.distributions, torch.nn.init
import torch.optim

# Network that learns the expected reward from a state.
class ValueCritic(nn.Module):
    def __init__(self, neural_module, values=1):
        super(ValueCritic, self).__init__()
        self.neural_module = neural_module
        self.input_dimension = list(more_itertools.always_iterable(neural_module.output_dimension))
        self.__input_size = functools.reduce(lambda x,y: x*y, self.input_dimension, 1)
        self.output_dimension = values
        self.output_layer = torch.nn.Linear(self.__input_size, values) # type: ignore

        for x in self.parameters():
            if x.dim() > 1:
                nn.init.kaiming_normal_(x)

    def recurrent(self):
        return self.neural_module.recurrent()

    def save_hidden(self):
        assert self.recurrent()
        return self.neural_module.save_hidden()


    def restore_hidden(self, state=None):
        assert self.recurrent()
        return self.neural_module.restore_hidden(state)


    def forward(self, input):
        output = self.neural_module(input)
        output = self.output_layer(output)
        return output



import torch

# Given a buffer of actions and policies, compute the log probability of choosing action[t+1] under policy[t]
def old_log_probs(action_buffer, policy_buffer, device):
    logprob_old = torch.zeros([*policy_buffer.shape], dtype=torch.float32, device=device) # type: ignore
    max_t = policy_buffer.shape[-1] - 1
    # TODO: Figure out what the bootstrap value should be
    logprob_old[max_t] = policy_buffer[max_t].log_prob(action_buffer[max_t]).sum()
    for t in range(max_t):
        logprob_old[t+1] = policy_buffer[t].log_prob(action_buffer[t+1]).sum()
    return logprob_old

# Compute the temporal difference residual of the rewards.
def td_residual(reward_buffer, estimated_values, gamma):
    td = torch.zeros([*reward_buffer.shape], dtype=torch.float32, device=reward_buffer.device) # type: ignore
    max_t = reward_buffer.shape[-1] - 1
    assert 0 and "This hasn't been tested yet."
    td[max_t] = estimated_values[max_t] -  reward_buffer[max_t]
    td[0:-1] = reward_buffer + gamma*estimated_values[1:] - estimated_values[:-1]
    return td

# Compute the generalized advantage estimation of our model.
# Caller must compute estimated values outside of this function.
def gae(reward_buffer, estimated_values, gamma, lambd=1):
    advantage = torch.zeros([*reward_buffer.shape], dtype=torch.float32, device=reward_buffer.device) # type: ignore
    max_t = reward_buffer.shape[-1] - 1
    # TODO: Determine how to predict "last" advantage value
    previous, advantage[max_t] = 0, estimated_values[max_t] -  reward_buffer[max_t]
    for t in range(max_t - 1, -1, -1):
        reward = (reward_buffer[t]
                + gamma*estimated_values[t+1] - estimated_values[t])
        previous = advantage[t] = reward + lambd*gamma*previous
    return advantage

# Compute (in)finite horizion discounted rewards.
# Uses discount factor of gamma, and multiplies lambd(a) by the current element.
# May cause excessively deep gradients, but computes in linear time.	
def discounted_returns(reward_buffer, gamma):
    assert 0.0 < gamma and gamma <= 1.0
    discounted_rewards = torch.zeros([*reward_buffer.shape], dtype=torch.float32, device=reward_buffer.device) # type: ignore
    
    # By iterating backwards over the rewards, we are computing a linear filter on the output.
    # That is: O[t] = I[t] + gamma*O[t+1], where I is the input array and O is the output array.
    # This works by "splitting off" the first term of the summation at t'=t, and then reusing the already-computer sum
    # from t'=t+1..T-1.
    # This approach may accumulate larger roundoff errors when t ≪ stop_time.
    previous = 0
    for t in range(reward_buffer.shape[0]-1, -1, -1):
        previous =  discounted_rewards[t] = reward_buffer[t] + gamma*previous
    return discounted_rewards




import torch
import numpy as np

import librl.utils

# Enapsulate all replay memory of a single task
# TODO: Allow data to be moved to GPU, and make device required.
class BaseEpisode:
    def __init__(self, obs_space, act_space, episode_length=200, device='cpu'):
        self.state_buffer =  []
        self.action_buffer = []
        self.logprob_buffer = torch.zeros([episode_length], dtype=torch.float32).to(device) # type: ignore
        self.reward_buffer = torch.zeros([episode_length], dtype=torch.float32).to(device) # type: ignore
        self.policy_buffer = np.full([episode_length], None, dtype=object)
        self.done =  None

    def log_state(self, t, state):
        self.state_buffer[t] = state
    def log_action(self, t, action, logprob):
        self.action_buffer[t] = action
        self.logprob_buffer[t] = logprob
    def log_rewards(self, t, reward):
        self.reward_buffer[t] = reward
    def log_policy(self, t, policy):
        self.policy_buffer[t]= policy
    def log_done(self, t):
        self.done = t

    def clear_replay(self):
        map(lambda x: x.fill_(0).detach_(), [self.logprob_buffer, self.reward_buffer])
        self.policy_buffer.fill(None)
        self.done = None

# Episode where action, state are torch contiguous and simple.
# Therefore, they are torch tensors.
class BoxEpisode(BaseEpisode):
    def __init__(self, obs_space, act_space, episode_length=200, device='cpu'):
        super(BoxEpisode, self).__init__(obs_space, act_space, episode_length=episode_length, device=device)
        self.state_buffer = torch.zeros([episode_length, *obs_space.shape], dtype=librl.utils.convert_np_torch(obs_space.dtype)).to(device) # type: ignore
        self.action_buffer = torch.zeros([episode_length, *act_space.shape], dtype=librl.utils.convert_np_torch(act_space.dtype)).to(device) # type: ignore
    def clear_replay(self):
        super(BoxEpisode, self).clear_replay()
        map(lambda x: x.fill_(0).detach_(), [self.state_buffer, self.action_buffer])
        

# Episode where action, state are cartesian products of tensors
# Therfore, we must store these elements in lists rather than tensor.
class ProductEpisode(BaseEpisode):
    def __init__(self, obs_space, act_space, episode_length=200, device='cpu'):
        super(ProductEpisode, self).__init__(obs_space, act_space, episode_length=episode_length, device=device)
        self.state_buffer = np.full([episode_length], None, dtype=object)# type: ignore
        self.action_buffer = np.full([episode_length], None, dtype=object)# type: ignore
    def clear_replay(self):
        super(ProductEpisode, self).clear_replay()
        map(lambda x: x.fill(0), [self.state_buffer, self.action_buffer])


import functools

import torch
# Sample logging functions.
##############
# CC Loggers #
##############
def cc_reward_logger(epoch, task_samples):
    rewards = len(task_samples) * [None]
    for idx, task in enumerate(task_samples):
        rewards[idx] = sum(task.trajectories[0].reward_buffer.view(-1))

    mean_reward = (sum(rewards)/len(rewards)).item() # type: ignore
    print(f"R^bar_({epoch}) = {mean_reward}.")

# Useful logger for robotics where actions' magitudes really matter.
def cc_action_reward_logger(epoch, task_samples):
    rewards, mu_act = len(task_samples) * [None],  len(task_samples) * [None]
    for idx, task in enumerate(task_samples):
        mu_act[idx] = torch.mean(task.trajectories[0].action_buffer.type(torch.float32), (0)) # type: ignore
        rewards[idx] = sum(task.trajectories[0].reward_buffer.view(-1))

    mean_reward = (sum(rewards)/len(rewards)).item() # type: ignore
    mean_action = functools.reduce(lambda x, y: x+y, mu_act, 0).mean()
    max_action = functools.reduce(lambda x, y: torch.max(x.abs(), y.abs()), mu_act)
    print(f"R^bar_({epoch}) = {mean_reward} with {mean_action:.4f} {max_action.data}.")

#########################
# Classification Loggers#
#########################
# Logger that records per-task accuracy each epoch
def cls_accuracy_logger(epoch, task_samples, correct_total_pairs):
    correct, total = correct_total_pairs
    print(correct, total)
    for idx, _ in enumerate(task_samples):
        print(f"Accuracy on task {idx} of {correct[idx]/total[idx]}")


import functools

import more_itertools
import torch
import torch.nn as nn
import torch.optim


# A simple NN designed for classification.
# It performs not output normalization. 
class Classifier(nn.Module):
    def __init__(self, neural_module, output_dimension):
        super(Classifier, self).__init__()
        self.neural_module = neural_module
        self.input_dimension = list(more_itertools.always_iterable(neural_module.output_dimension))

        self.output_dimension = output_dimension
        self.__input_size = functools.reduce(lambda x,y: x*y, self.input_dimension, 1)
        self.output_layer = nn.Linear(self.__input_size, self.output_dimension)
        self.optimizer = torch.optim.Adam(self.parameters(), lr=0.003, weight_decay=0.1)
    
    def recurrent(self):
        return self.neural_module.recurrent()

    def save_hidden(self):
        assert self.recurrent()
        return self.neural_module.save_hidden()

    def restore_hidden(self, state=None):
        assert self.recurrent()
        self.neural_module.restore_hidden(state)

    def forward(self, inputs):
        # Require that outputs be of the shape specified by our neural module.
        outputs = self.neural_module(inputs).view(-1, *self.input_dimension)
        # Then cast it to something usable by a MLP.
        outputs = outputs.view(-1, self.__input_size)
        outputs = self.output_layer(outputs)

        return outputs




def add_agent_attr(policy_based=False, allow_callback=False, allow_update=False):
    def deco(cls):
        attrs = {}
        attrs['policy_based'] = policy_based
        attrs['allow_callback'] = allow_callback
        attrs['allow_update'] = allow_update
        # By default, a model should not be recurrent.
        attrs['recurrent'] = lambda x: False
        for attr in attrs:
            setattr(cls,attr, attrs[attr])
        return cls
    return deco


import gym
import numpy as np
import torch

import librl.replay
# The current module will be imported before librl.task is finished, so we
# will need to refer to our parent class by relative path.
from .base import Task as _Task

class ContinuousControlTask(_Task):
    # Aliase batch size as the episode count.
    episode_count=property(_Task.batch_size, _Task.batch_size)

    # Methods to fill a task's sequential replay buffer.
    @staticmethod
    def sample_trajectories(task):
        task.clear_trajectories()
        task.init_env()

        # Massage something that looks like a tensor, ndarray to a tensor on the correct device.
        # If given an iterable of those things, recurssively attempt to massage them into tensors.
        def torchize(maybe_tensor):
            # Sometimes an object is a torch tensor that just needs to be moved to the right device
            if torch.is_tensor(maybe_tensor): return maybe_tensor.to(task.device)
            # Sometimes it is an ndarray
            elif isinstance(maybe_tensor, np.ndarray): return torch.tensor(maybe_tensor).to(task.device) # type: ignore
            # Maybe a simple ierable of things we need to move to torch, like a cartesian product of ndarrays.
            elif isinstance(maybe_tensor, (list, tuple)): return [torchize(item) for item in maybe_tensor]
            else: raise NotImplementedError(f"Don't understand the datatype of {type(maybe_tensor)}")

        for i in range(task.trajectory_count):
            state = torchize(task.env.reset()) # type: ignore
            episode = task.replay_ctor(task.env.observation_space, task.env.action_space, task.episode_length)
            episode.log_done(task.episode_length + 1)
            for t in range(task.episode_length):
                
                episode.log_state(t, state)

                action, logprob_action = task.agent.act(state)
                episode.log_action(t, action, logprob_action)
                if task.agent.policy_based: episode.log_policy(t, task.agent.policy_latest)
                # If our action is a tensor, it needs to be migrated to the CPU for the simulator.
                if torch.is_tensor(action): action = action.view(-1).detach().cpu().numpy()
                state, reward, done, _ = task.env.step(action)
                
                if task.agent.allow_callback: task.agent.act_callback(state=state, reward=reward)
                state, reward = torchize(state), torch.tensor(reward).to(task.device) # type: ignore
                episode.log_rewards(t, reward)
                if done: 
                    episode.log_done(t+1)
                    break

            task.add_trajectory(episode)

    # Use sample_trajectories as the default sample, unless otherwise specified.
    def __init__(self, sample_fn = sample_trajectories.__func__, replay_ctor=librl.replay.episodic.BoxEpisode, 
    env=None, agent=None, trajectories=1, episode_length=100):

        super(ContinuousControlTask,self).__init__(librl.task.ProblemTypes.ContinuousControl)
        assert env is not None and agent is not None
        assert isinstance(env, gym.Env)

        self.env = env
        self.agent = agent
        self._episode_length = episode_length
        self.trajectory_count = trajectories

        self.sample = sample_fn
        self.replay_ctor = replay_ctor
        self.trajectories = []

    # Override in subclass!!
    def init_env(self):
        raise NotImplemented("Please implement this method in your subclass.")

    def add_trajectory(self, trajectory):
        self.trajectories.append(trajectory)
    def clear_trajectories(self):
        self.trajectories = []

    # Let the task control the sampled batch size
    @property
    def episode_length(self):
        return self._episode_length
    @episode_length.setter
    def episode_length(self, value):
        self._episode_length = value

# Example class that allows us to run gym tasks with no extra effort.
class ContinuousGymTask(ContinuousControlTask):
    def __init__(self, **kwargs):
        super(ContinuousGymTask, self).__init__(**kwargs)
    def init_env(self):
        # Most gym environment need no extra init'ing
        pass


import functools

import more_itertools
import torch
import torch.nn as nn
import torch.distributions, torch.nn.init
import torch.optim


import librl.nn.policy

# Agent network based on a submission to my COSC689 class
# It is a stochastic policy network. It will return the policy from forward,
# and you can use this policy to generate further samples
# The current policy is sampling random values from a torch Categorical distrubtion
# conditioned on the output of a linear network.
# The Categorical distribution is non-differentiable, so this may cause
# problems for future programmers.
class BiCategoricalActor(nn.Module):
    def __init__(self, neural_module, action_space, observation_space):
        super(BiCategoricalActor, self).__init__()
        
        self.input_dimension = list(more_itertools.always_iterable(neural_module.output_dimension))
        self.__input_size = functools.reduce(lambda x,y: x*y, self.input_dimension, 1)
        self.neural_module = neural_module
        self.output_dimension = 2
        assert len (action_space.shape) == 2 and action_space.shape[-1] == 2
        assert len (observation_space.shape) == 2 and observation_space.shape[0] == observation_space.shape[1]

        # Our output layers are used as the seed for some set of random number generators.
        # These random number generators are used to generate edge pairs.
        self.output_layers = {}
        self.output_layers["first"] = nn.Linear(self.__input_size, observation_space.shape[0])
        self.output_layers["second"] = nn.Linear(self.__input_size, observation_space.shape[0])
        self.output_layers = nn.ModuleDict(self.output_layers)

        # Must pass output layers through softmax in order for them to be a proper PDF.
        self.softmax = nn.Softmax(dim=0)

        # Initialize NN
        for x in self.parameters():
            if x.dim() > 1:
                nn.init.kaiming_normal_(x)
                
    def recurrent(self):
        return self.neural_module.recurrent()

    def save_hidden(self):
        assert self.recurrent()
        self.neural_module.save_hidden()

    def restore_hidden(self, state=None):
        assert self.recurrent()
        self.neural_module.restore_hidden(state)

    def forward(self, input):
        output = self.neural_module(input).view(-1, self.__input_size)
        actions = []

        assert not torch.isnan(output).any() # type: ignore

        # Treat the outputs of my softmaxes as the probability distribution for my NN.
        first_preseed = self.output_layers["first"  ](output).view(-1)
        second_preseed = self.output_layers["second"](output).view(-1)

        # Since softmax will rescale all numbers to sum to 1,
        # logically it doesn't matter where the sequence lies on the number line.
        # Since softmax will compute e(x), I want all x's to be small.
        # To do this, I subtract the maximum value from every element, moving my
        # numbers from (-∞,∞) to (-∞,0]. This makes softmax more stable
        first_seed = first_preseed - torch.max(first_preseed) # type: ignore
        second_seed = second_preseed - torch.max(second_preseed) # type: ignore
        first_seed = self.softmax(first_seed)
        second_seed = self.softmax(second_seed)
        
        # Encapsulate our poliy in an object so downstream classes don't
        # need to know what kind of distribution to re-create.
        policy = librl.nn.policy.BiCategoricalPolicy(first_seed, second_seed)
        # Sample edge pair to toggle
        actions = policy.sample((1,))
        # Each actions is drawn independtly of others, so joint prob
        # is all of them multiplied together. However, since we have logprobs,
        # we need to sum instead.
        log_prob = torch.sum(policy.log_prob(actions)) # type: ignore
        # Arrange actions so they look like actions from other models.

        return actions, log_prob, policy



class IndependentNormalActor(nn.Module):
    def __init__(self, neural_module, action_space, observation_space, policy_ctor=librl.nn.policy.RepeatedNormal):
        super(IndependentNormalActor, self).__init__()
        self.policy_ctor = policy_ctor

        self.input_dimension = list(more_itertools.always_iterable(neural_module.output_dimension))
        self.__input_size = functools.reduce(lambda x,y: x*y, self.input_dimension, 1)
        self.neural_module = neural_module
        self.output_dimension = action_space.shape
        self.__output_size = functools.reduce(lambda x,y: x*y, self.output_dimension, 1)

        # Our output layers are used as the seed for some set of random number generators.
        # These random number generators are used to generate edge pairs.
        self.mu_layer = nn.Linear(self.__input_size, self.__output_size)
        self.cov_diag = nn.Linear(self.__input_size, self.__output_size)
        self.make_sane = torch.nn.Softsign() # type: ignore

        # Initialize NN
        for x in self.parameters():
            if x.dim() > 1:
                nn.init.kaiming_normal_(x)

    def recurrent(self):
        return self.neural_module.recurrent()
        
    def save_hidden(self):
        assert self.recurrent()
        return self.neural_module.save_hidden()

    def restore_hidden(self, state=None):
        assert self.recurrent()
        self.neural_module.restore_hidden(state)

    def forward(self, input):
        output = self.neural_module(input).view(-1, self.__input_size)

        # Treat the output of NN as seed of a mu network.
        mu = self.mu_layer(output)

        # Variance must be +'ve, and softplus does this nicely.
        # For other activation ideas, see:
        #    https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity
        var = (torch.nn.Softsign()(self.cov_diag(output)) + 1)/2 # type: ignore

        # Encapsulate our poliy in an object so downstream classes don't
        # need to know what kind of distribution to re-create.
        policy = self.policy_ctor(mu, var)
        # Cheating normal, average quickly becomes 0.
        # RepeatedNormal w/sofptlus, abs(average action) -> bigger than 1.
        # Right now, average action is a function(# epochs), that's bad.
        # Maybe l2 is too high (see utils.py), or dropout is too high. Over regularized.

        actions = policy.sample((1,))
        
        # Each actions is drawn independtly of others, so joint prob
        # is all of them multiplied together. However, since we have logprobs,
        # we need to sum instead.
        #print(var, policy.log_prob(actions))
        log_prob = torch.sum(policy.log_prob(actions)) # type: ignore

        return actions.view(*self.output_dimension), log_prob, policy



from .cnn import ConvolutionalKernel
from .mlp import MLPKernel
from .rnn import RecurrentKernel
from .sequential import SequentialKernel
from .bilinear import BilinearKernel, SiameseKernel, JoinKernel


import functools
import types

import more_itertools
import torch
import torch.nn as nn
import torch.optim

# Forward takes in two inputs, and passes them through two internal neural modules.
# It combine the results via a single bilinear layer.
# Can extract realtionships 
class BilinearKernel(nn.Module):
    def __init__(self, left_module, right_module, output_size, nlo = None):
        super(BilinearKernel, self).__init__()

        self.nlo = nlo if nlo != None else lambda x: x

        self.__left_module = left_module
        self.__right_module = right_module


        self.output_dimension = (output_size, )

        self.__left_output__size = functools.reduce(lambda x, y: x*y, left_module.output_dimension, 1)
        self.__right_output__size = functools.reduce(lambda x, y: x*y, right_module.output_dimension, 1)
        self.bilinear_layer = nn.Bilinear(self.__left_output__size, self.__right_output__size, output_size)

        # Initialize NN
        for x in self.parameters():
            if x.dim() > 1:
                nn.init.kaiming_normal_(x)

                
    def recurrent(self):
        return self.__left_module.recurrent() or self.__right_module.recurrent()

    def save_hidden(self):
        ret = {}
        assert self.recurrent()
        if self.__left_module.recurrent(): ret[id(self.__left_module)] = self.__left_module.save_hidden()
        # If the both networks are the same, don't save it twice.
        if self.__left_module == self.__right_module: pass
        elif self.__right_module.recurrent(): ret[id(self.__right_module)] = self.__right_module.save_hidden()
        return ret
            
    def restore_hidden(self, state_dict=None):
        if self.__left_module.recurrent():
            id_left = id(self.__left_module)
            if state_dict != None and id_left in state_dict: self.__left_module.restore_hidden(state_dict[id_left])
            else: self.__left_module.restore_hidden()
        # If the both networks are the same, don't restore twice
        if self.__left_module == self.__right_module: pass
        elif self.__right_module.recurrent():
            id_right = id(self.__right_module)
            if state_dict != None and id_right in state_dict: self.__right_module.restore_hidden(state_dict[id_right])
            else: self.__right_module.restore_hidden()

    def forward(self, linput, rinput):
        loutput = self.__left_module(linput).view(-1, self.__left_output__size)
        routput = self.__right_module(rinput).view(-1, self.__right_output__size)

        output = self.bilinear_layer(loutput, routput)
        output = self.nlo(output)
        assert not torch.isnan(output).any() # type: ignore

        return output

# Share one network between two inputs.
# The supplied module is responsible for handling various data sizes.
# Inspired by Koch et al. 2015 `Siamese Neural Networks for One-shot Image Recognition`
class SiameseKernel(BilinearKernel):
    def __init__(self, module, output_size, nlo = torch.nn.Sigmoid()):
        super(SiameseKernel, self).__init__(module, module, output_size, nlo=nlo)

# Feed a single input through two different neural modules.
# Join the result using bilinear layer
class JoinKernel(BilinearKernel):
    def __init__(self, left_module, right_module, output_size, nlo = None):
        super(JoinKernel, self).__init__(left_module, right_module, output_size, nlo=nlo)
    def forward(self, input):
        return BilinearKernel.forward(self, input, input)


import functools
import types

import more_itertools
import torch
import torch.nn as nn
import torch.optim

class RecurrentKernel(nn.Module):
    def __init__(self, input_dimensions, hidden_size, num_layers, recurrent_unit="LSTM", bidirectional=False, dropout=None):
        super(RecurrentKernel, self).__init__()
        dropout = dropout if dropout else self.get_default_hyperparameters().dropout
        self.hidden_size = hidden_size
        self.num_layers = num_layers

        self.input_dimensions = list(more_itertools.always_iterable(input_dimensions))
        self.__input__size = functools.reduce(lambda x, y: x*y, self.input_dimensions, 1)
        self.output_dimension = (hidden_size, )

        
        if recurrent_unit.upper() == "LSTM": rec_init = nn.LSTM
        elif recurrent_unit.upper() == "GRU": rec_init = nn.LSTM
        elif recurrent_unit.upper() == "RNN": rec_init = nn.LSTM
        else:rec_init = lambda *x: (_ for _ in ()).throw(NotImplementedError("Choose an implemented recurrent unit"))
        
        self.recurrent_layer = rec_init(self.__input__size, hidden_size, num_layers = num_layers, bidirectional=bidirectional, batch_first=False)
        self.init_hidden()
        # Initialize NN
        for x in self.parameters():
            if x.dim() > 1:
                nn.init.kaiming_normal_(x)

    def init_hidden(self):

        self.hidden_state = torch.zeros(self.num_layers, 1, self.hidden_size) # type: ignore
        self.cell_state = torch.zeros(self.num_layers, 1, self.hidden_size) # type: ignore
        nn.init.kaiming_normal_(self.hidden_state)
        nn.init.kaiming_normal_(self.cell_state)

    def recurrent(self):
        return True

    def save_hidden(self):
        return self.hidden_state

    def restore_hidden(self, state=None):
        if state == None:
            self.init_hidden()
        else:
            assert 0
            hidden_state, cell_state = state
            # Assert that state be correct shape for calling forward().
            # Assert here rather than forward(), since it may not be obvious why forward fails
            # when cell state is of wrong shape.
            hidden_state = hidden_state.view(self.num_layers, -1, self.hidden_size)
            cell_state = cell_state.view(self.num_layers, -1, self.hidden_size)

            self.hidden_state, self.cell_state = hidden_state, cell_state

    @staticmethod
    def get_default_hyperparameters():
        ret = {}
        ret['dropout'] = .1
        ret['layer_list'] = [200, 100]
        ret['recurrent_unit'] = 'RNN'

        return types.SimpleNamespace(**ret)

    def forward(self, input):
        # Treat the entire input as a single seqence of data.
        d0, d1 = 1, -1
        if len(input.shape) == 1: assert self.__input__size == input.shape[-1]
        elif len(input.shape) == 2: 
            assert self.__input__size == input.shape[-1]
            d0 = input.shape[0]
        elif len(input.shape) == 3: 
            raise NotImplementedError("No clue how to do this.")
        # TODO: Figure out how to batch hidden states
        input = input.view(d0, d1, self.__input__size)
        # Push observations through feed forward layers.
        if self.hidden_state.shape[1] == 1 and input.shape[1] != 1:
            hidden_state = self.hidden_state.repeat((1,input.shape[1],1))
        else: hidden_state = self.hidden_state

        if self.cell_state.shape[1] == 1 and input.shape[1] != 1:
            cell_state = self.cell_state.repeat((1,input.shape[1],1))
        else: cell_state = self.cell_state

        h0 = hidden_state, cell_state
        output, h1 = self.recurrent_layer(input.float(), h0)

        self.hidden_state, self.cell_state = h1
        # We really dont care about the progress / history of our state data.
        self.hidden_state, self.cell_state = self.hidden_state.detach(), self.cell_state.detach()
        assert not torch.isnan(output).any() # type: ignore

        return output


import functools

import torch

from .log import *


def cc_episodic_trainer(train_info, task_dist, train_fn, log_fn = cc_reward_logger):
    for epoch in range(train_info['epochs']):
        task_samples = task_dist.sample(train_info['task_count'])
        train_fn(task_samples)
        log_fn(epoch, task_samples)

def cls_trainer(train_info, task_dist, train_fn, log_fn = cls_accuracy_logger):
    for epoch in range(train_info['epochs']):
        task_samples = task_dist.sample(train_info['task_count'])
        result = train_fn(task_samples)
        log_fn(epoch, task_samples, result)


import functools

import torch

import librl.calc
import librl.replay


@functools.singledispatch
def _estimate_values(trajectory, critic_fn, shape=None):
    raise NotImplementedError("I don't know how to handle this trajectory type!")

@_estimate_values.register(librl.replay.episodic.BoxEpisode)
def _(trajectory, critic_fn, shape=None):
    states = trajectory.state_buffer[:trajectory.done]
    shape = shape if shape else (-1, *(states.shape[1:]))
    states = states.view(*shape)
    return critic_fn(states)

@_estimate_values.register(librl.replay.episodic.ProductEpisode)
def _(trajectory, critic_fn, shape=None):
    # TODO: Respect shape if present.
    states = trajectory.state_buffer[:trajectory.done]
    out = torch.zeros(trajectory.reward_buffer.shape).to(trajectory.reward_buffer.device)
    for idx, state in enumerate(states):
        v = critic_fn(state)
        out[idx] = v 
    return out

###################################
# Exploration Bonuses Computation #
###################################

# Compute the entropy of a trajectory, which is just p(x)log(p(x))
class basic_entropy_bonus:
    def __init__(self, beta=0.01):
        self.beta = beta
    def __call__(self, trajectory):
        return -self.beta * trajectory.logprob_buffer[:trajectory.done] * trajectory.logprob_buffer[:trajectory.done].exp()

##############################
# Reward Discounting Methods #
##############################
# Different reward functions taken from:
#   https://danieltakeshi.github.io/2017/04/02/notes-on-the-generalized-advantage-estimation-paper/

# The reward at each timestep is the sum of all rewards over the epoch
def total_reward(trajectory):
    return torch.full(*trajectory.reward_buffer.shape, torch.sum(trajectory.reward_buffer)) # type: ignore

# At each timestep, compute the infinite-horizon discounted reward.
class to_go_reward:
    def __init__(self, gamma=.975):
        self.gamma = gamma
    def __call__ (self, trajectory):
        return librl.calc.discounted_returns(trajectory.reward_buffer[:trajectory.done], gamma=self.gamma)

# At each timestep, compute the infinite-horizon discounted reward and subtract a baseline generate by a critic nn.
class baseline_to_go(to_go_reward):
    def __init__(self, critic_fn, gamma=.975):
        super(baseline_to_go, self).__init__(gamma=gamma)
        self.critic_fn = critic_fn
    def __call__ (self, trajectory):
        with torch.no_grad(): estimated_values = _estimate_values(trajectory, self.critic_fn).view(-1)
        return super().__call__(trajectory) - estimated_values

# Compute the temporal difference residual of the rewards.
class td_residual:
    def __init__(self, critic_fn, gamma=.975):
        self.critic_fn = critic_fn
        self.gamma = gamma
    def __call__(self, trajectory):
        with torch.no_grad(): estimated_values = estimated_values = _estimate_values(trajectory, self.critic_fn).view(-1)
        return librl.calc.td_residual(trajectory.reward_buffer[:trajectory.done], estimated_values, gamma=self.gamma)



import numpy as np
from numpy.random import Generator, PCG64

# Class representing a weighted task distribution.
# You can sample a number of tasks from this with replacements
class TaskDistribution:
    def __init__(self, seed=None):
        self._tasks = []
        self._task_weights = []
        self._rng = Generator(PCG64(seed))

    def add_task(self, task, weight=1):
        self._tasks.append(task)
        self._task_weights.append(weight)

    def sample(self, count):
        # Must convert weights to proabbilities.
        mean = sum(self._task_weights)
        task_probabilities = list(map(lambda x: x/mean, self._task_weights))
        # Task definitions are different than task objects.
        # If they were the same, and we sampled on tasks directly,
        # it was possible to return the same task object twice.
        # This lead to accidentally overwriting replay memory.
        definitions = self._rng.choice(self._tasks, size=(count,), replace=True, p=task_probabilities)
        return [definition.instance() for definition in definitions]
        
    def clear_tasks(self):
        self._tasks = []
        self._task_weights = []





import itertools

import numpy as np
import torch
import torch.utils.data

# Used to split a single dataset over multple data loaders.
# It's the responsibility of the *caller* to ensure that each item appears only once in the dataset,
# otherwise elements may be in multiple splits.
def load_split_data(dset:torch.utils.data.dataset.Dataset, batch_size, loader_count, **dataloader_kwargs):
    assert batch_size >= 1 and loader_count >= 1
    part_size = len(dset) // loader_count
    # Make even sized batches for the first n-1 loaders, and stick remaining elements in last
    head = (loader_count-1)*(part_size,)
    sizes = *head, len(dset)-sum(head),
    #for i in sizes: print(i)
    # Make a data loader for each dataset
    return [torch.utils.data.DataLoader(i, batch_size=batch_size, shuffle=True, **dataloader_kwargs) for i in torch.utils.data.random_split(dset, sizes, None)] 

def convert_np_torch(item):
    if item in convert_np_torch._info: return convert_np_torch._info[item]
    elif item == np.float32: return torch.float32 # type: ignore
    elif item == np.int8: return torch.int8 # type: ignore
    elif item == np.int16: return torch.int16 # type: ignore
    print(f"Dtype ({item}) not implemented.")
    assert 0 and "Dtype not implemented."

convert_np_torch._info = {
        np.bool       : torch.bool, # type: ignore
        np.uint8      : torch.uint8, # type: ignore
        np.int8       : torch.int8, # type: ignore
        np.int16      : torch.int16, # type: ignore
        np.int32      : torch.int32, # type: ignore 
        np.int64      : torch.int64, # type: ignore
        np.float16    : torch.float16, # type: ignore
        np.float32    : torch.float32, # type: ignore
        np.float64    : torch.float64, # type: ignore
        np.complex64  : torch.complex64, # type: ignore
        np.complex128 : torch.complex128} # type: ignore




from distutils.core import setup
from setuptools import find_packages
import pkg_resources
install_reqs = []
install_reqs.extend(["more-itertools", "overrides", "future", "matplotlib", "scipy", "Pillow"])
install_reqs.extend(["networkx ~= 2.5", "torch ~= 1.6.0", "torchvision ~= 0.7.0"])
setup(name='librl',
      version='0.2.1',
      description='Library for various DRL and DL projects.',
      author='Matthew McRaven',
      author_email='mkm302@georgetown.edu',
      url='https://github.com/Matthew-McRaven/librl',
      install_reqs=install_reqs,
      python_requires='~= 3.8',
      packages=find_packages()
     )


import torch
import torch.distributions, torch.nn.init
import torch.optim
from overrides import overrides, EnforceOverrides

import librl.calc
import librl.reward
import librl.task
##########################
# Policy Gradient Losses #
##########################

# Compute the loss of a multi-trajecory task.
class PolicyLoss(EnforceOverrides):
    def __init__(self):
        super(PolicyLoss, self).__init__()

    # Override this method to compute the loss of a single trajectory.
    # Should return a single value.
    def compute_trajectory_loss(self, trajectory):
        raise NotImplementedError("Implement in subclass")

    # When called with a CC task, run the loss algorithm over each trajectory in the task.
    # This will implicitly perform the outer mean over the the number of trajectories,
    # and defer the computation of per-trajectory loss to a subclass via compute_trajectory_loss.
    def __call__(self, task):
        assert isinstance(task, librl.task.ContinuousControlTask)
        losses = []
        for trajectory in task.trajectories: losses.append(self.compute_trajectory_loss(trajectory))
        return sum(losses) / len(losses)

# For any class base on log_prob*reward, this class is convenient.
# Pass in a function(trajectory)->array[rewards], and this will do the rest.
# Especially useful for VPG, PGB
class LogProbBased(PolicyLoss):
    def __init__(self, reward_fn, explore_bonus_fn=lambda _:0):
        super(LogProbBased, self).__init__()
        assert callable(reward_fn) and callable(explore_bonus_fn)
        self.reward_fn = reward_fn
        self.explore_bonus_fn = explore_bonus_fn
    
    @overrides # type: ignore
    def compute_trajectory_loss(self, trajectory):
        return sum(trajectory.logprob_buffer[:trajectory.done] * self.reward_fn(trajectory) + self.explore_bonus_fn(trajectory))

class VPG(LogProbBased):
    def __init__(self, gamma=0.975, explore_bonus_fn=lambda _: 0): 
        assert callable(explore_bonus_fn)
        super(VPG, self).__init__(librl.reward.to_go_reward(gamma=gamma), explore_bonus_fn=explore_bonus_fn)
class PGB(LogProbBased):
    def __init__(self, critic_fn, gamma=0.975, explore_bonus_fn=lambda _: 0): 
        assert callable(explore_bonus_fn)
        super(PGB, self).__init__(librl.reward.baseline_to_go(critic_fn, gamma=gamma), explore_bonus_fn=explore_bonus_fn)

# Proximal policy optimization update / loss function.
class PPO(PolicyLoss):
    def __init__(self,  critic_fn, gamma=0.975, lambd=.99, epsilon=.2, c_1=.5, explore_bonus_fn = lambda _: 0):
        assert callable(explore_bonus_fn)
        self.critic_fn = critic_fn
        self.gamma = gamma
        self.lambd = lambd
        self.epsilon = epsilon
        self.c_1 = c_1
        self.explore_bonus_fn = explore_bonus_fn

    @overrides # type: ignore
    def compute_trajectory_loss(self, trajectory):
        device = trajectory.reward_buffer.device
        # Don't propogate gradients into critic when updating actor.
        with torch.no_grad(): estimated_values = librl.reward._estimate_values(trajectory, self.critic_fn)
        # Augment rewards with a bonus for exploration, like policy entropy.
        augmented_rewards = trajectory.reward_buffer[:trajectory.done] + self.explore_bonus_fn(trajectory)
        discounted = librl.calc.discounted_returns(augmented_rewards, gamma=self.gamma)
        A =  librl.calc.gae(trajectory.reward_buffer[:trajectory.done], estimated_values, self.gamma)
        log_prob_old = librl.calc.old_log_probs(trajectory.action_buffer[:trajectory.done], trajectory.policy_buffer[:trajectory.done], device)
        # Compute indiviudal terms of the PPO algorithm.
        ratio = trajectory.logprob_buffer[:trajectory.done].exp() / (log_prob_old.exp() + 1e-6)
        lhs, rhs = ratio * A, torch.clamp(ratio, 1-self.epsilon, 1+self.epsilon) * A # type: ignore
        minterm = torch.min(lhs, rhs) # type: ignore
        subterm = self.c_1 * (discounted-estimated_values.view(-1)).pow(2)
        return torch.sum(minterm - subterm) # type: ignore


import more_itertools
import torch

import librl.task
# The current module will be imported before librl.task is finished, so we
# will need to refer to our parent class by relative path.
from .base import Task as _Task

# Given a classifier and a dataset, the goal is to train the classifier to predict dataset labels
# with high portability to the testing data.
# train_data_iter, validation_data_iter should be an iterable of DataLoaders rather than bare DataLoaders.
class ClassificationTask(_Task):
    def __init__(self, classifier=None, criterion=None, train_data_iter=None, validation_data_iter=None, train=True,**kwargs):
        super(ClassificationTask, self).__init__(librl.task.ProblemTypes.Classification, **kwargs)
        assert classifier is not None and criterion is not None
        assert (train_data_iter is not None) or (validation_data_iter is not None)
        # Train / validation data should be an iterable of data loaders rather than a single data loader.
        assert not isinstance(train_data_iter, torch.utils.data.DataLoader) # type: ignore
        assert not isinstance(validation_data_iter, torch.utils.data.DataLoader) # type: ignore

        self.classifier = classifier
        self.criterion = criterion
        self.train = train

        self.train_batch = train_one_batch
        self.train_data_iter = train_data_iter
        self.validate_batch = test_one_batch
        self.validation_data_iter = validation_data_iter

# Default implementation that trains a classifier on a single batch of data
def train_one_batch(classifier, criterion, data, target):
    output = classifier(data)
    loss = criterion(output, target)
    loss.backward()
    classifier.optimizer.step()
    classifier.optimizer.zero_grad()
    return loss.item(), output.argmax(dim=1)

# Default implementation that tests a classifier on a single batch of data.
def test_one_batch(classifier, criterion, data, target):
    output = classifier(data)
    loss = criterion(output, target)
    return loss.item(), output.argmax(dim=1)